@@ -23,6 +23,9 @@
 #include <uuid.h>
 #include <dconf.h>
 
+#define G_SETTINGS_ENABLE_BACKEND
+#include <gio/gsettingsbackend.h>
+
 #include "terminal-type-builtins.h"
 #include "terminal-schemas.h"
 #include "terminal-debug.h"
@@ -177,6 +180,16 @@ terminal_settings_list_valid_uuid (const char *str)
   return uuid_parse ((char *) str, u) == 0;
 }
 
+static gboolean
+settings_backend_is_dconf (void)
+{
+  gs_unref_object GSettingsBackend *backend;
+
+  backend = g_settings_backend_get_default ();
+
+  return g_str_equal (G_OBJECT_TYPE_NAME (backend), "DConfSettingsBackend");
+}
+
 static char *
 new_list_entry (void)
 {
@@ -267,9 +280,7 @@ clone_child (TerminalSettingsList *list,
   char **keys;
   guint i;
   gs_unref_object DConfClient *client;
-#ifndef HAVE_DCONF_1_2
   DConfChangeset *changeset;
-#endif
 
   new_uuid = new_list_entry ();
 
@@ -279,16 +290,12 @@ clone_child (TerminalSettingsList *list,
   path = path_new (list, uuid);
   new_path = path_new (list, new_uuid);
 
-#ifdef HAVE_DCONF_1_2
-  client = dconf_client_new (NULL, NULL, NULL, NULL);
-#else
   client = dconf_client_new ();
   changeset = dconf_changeset_new ();
-#endif
 
   /* FIXME: this is beyond ugly. Need API on GSettingsSchema to list all the keys! */
   {
-    gs_unref_object GSettings *dummy = g_settings_new_with_path (list->child_schema_id, "/foo");
+    gs_unref_object GSettings *dummy = g_settings_new_with_path (list->child_schema_id, "/foo/");
     keys = g_settings_list_keys (dummy);
   }
 
@@ -297,27 +304,17 @@ clone_child (TerminalSettingsList *list,
     gs_unref_variant GVariant *value;
 
     rkey = g_strconcat (path, keys[i], NULL);
-#ifdef HAVE_DCONF_1_2
-    value = dconf_client_read_no_default (client, rkey);
-#else
     value = dconf_client_read (client, rkey);
-#endif
 
     if (value) {
       gs_free char *wkey;
       wkey = g_strconcat (new_path, keys[i], NULL);
-#ifdef HAVE_DCONF_1_2
-      dconf_client_write (client, wkey, value, NULL, NULL, NULL);
-#else
       dconf_changeset_set (changeset, wkey, value);
-#endif
     }
   }
 
-#ifndef HAVE_DCONF_1_2
   dconf_client_change_sync (client, changeset, NULL, NULL, NULL);
   dconf_changeset_unref (changeset);
-#endif
 
   return new_uuid;
 }
@@ -329,7 +326,7 @@ terminal_settings_list_add_child_internal (TerminalSettingsList *list,
   char *new_uuid;
   gs_strfreev char **new_uuids;
 
-  if (uuid)
+  if (uuid && settings_backend_is_dconf ())
     new_uuid = clone_child (list, uuid);
   else
     new_uuid = new_list_entry ();
@@ -349,8 +346,6 @@ terminal_settings_list_remove_child_internal (TerminalSettingsList *list,
                                               const char *uuid)
 {
   gs_strfreev char **new_uuids;
-  gs_free char *path = NULL;
-  gs_unref_object DConfClient *client = NULL;
 
   _terminal_debug_print (TERMINAL_DEBUG_SETTINGS_LIST,
                          "%s UUID %s\n", G_STRFUNC, uuid);
@@ -368,15 +363,14 @@ terminal_settings_list_remove_child_internal (TerminalSettingsList *list,
     g_settings_set_string (&list->parent, TERMINAL_SETTINGS_LIST_DEFAULT_KEY, "");
 
   /* Now we unset all keys under the child */
-  path = path_new (list, uuid);
+  if (settings_backend_is_dconf ()) {
+    gs_free char *path;
+    gs_unref_object DConfClient *client;
 
-#ifdef HAVE_DCONF_1_2
-  client = dconf_client_new (NULL, NULL, NULL, NULL);
-  dconf_client_write (client, path, NULL, NULL, NULL, NULL);
-#else /* modern DConf */
-  client = dconf_client_new ();
-  dconf_client_write_sync (client, path, NULL, NULL, NULL, NULL);
-#endif
+    path = path_new (list, uuid);
+    client = dconf_client_new ();
+    dconf_client_write_sync (client, path, NULL, NULL, NULL, NULL);
+  }
 }
 
 static void
@@ -823,7 +817,7 @@ terminal_settings_list_remove_child (TerminalSettingsList *list,
  * @list: a #TerminalSettingsList
  * @child: a #GSettings of a child in the list
  *
- * Returns the UUID of @child in the list, or %NULL if @child is in the list.
+ * Returns the UUID of @child in the list, or %NULL if @child is not in the list.
  *
  * Returns: (transfer full): the UUID of the child in the settings list, or %NULL
  */
@@ -843,6 +837,7 @@ terminal_settings_list_dup_uuid_from_child (TerminalSettingsList *list,
   g_return_val_if_fail (p[0] == ':', NULL);
   p++;
   g_return_val_if_fail (strlen (p) == 37, NULL);
+  g_return_val_if_fail (p[36] == '/', NULL);
   p[36] = '\0';
   g_assert (terminal_settings_list_valid_uuid (p));
 
@@ -856,7 +851,7 @@ terminal_settings_list_dup_uuid_from_child (TerminalSettingsList *list,
  *
  * Sets @uuid as the default child.
  */
-void 
+void
 terminal_settings_list_set_default_child (TerminalSettingsList *list,
                                           const char *uuid)
 {
@@ -868,3 +863,43 @@ terminal_settings_list_set_default_child (TerminalSettingsList *list,
 
   g_settings_set_string (&list->parent, TERMINAL_SETTINGS_LIST_DEFAULT_KEY, uuid);
 }
+
+/**
+ * terminal_settings_list_foreach_child:
+ * @list: a #TerminalSettingsList
+ * @callback: a #TerminalSettingsListForeachFunc
+ * @user_data: user data for @callback
+ *
+ * Calls @callback for each child of @list.
+ *
+ * NOTE: No changes to @list must be made from @callback.
+ */
+void
+terminal_settings_list_foreach_child (TerminalSettingsList *list,
+                                      TerminalSettingsListForeachFunc callback,
+                                      gpointer user_data)
+{
+  g_return_if_fail (TERMINAL_IS_SETTINGS_LIST (list));
+  g_return_if_fail (callback);
+
+  for (char **p = list->uuids; *p; p++) {
+    const char *uuid = *p;
+    gs_unref_object GSettings *child = terminal_settings_list_ref_child_internal (list, uuid);
+    if (child != NULL)
+      callback (list, uuid, child, user_data);
+  }
+}
+
+/**
+ * terminal_settings_list_foreach_child:
+ * @list: a #TerminalSettingsList
+ *
+ * Returns: the number of children of @list.
+ */
+guint
+terminal_settings_list_get_n_children (TerminalSettingsList *list)
+{
+  g_return_val_if_fail (TERMINAL_IS_SETTINGS_LIST (list), 0);
+
+  return g_hash_table_size (list->children);
+}
